// MMSQ: 2048 15 2 256

/*
 * For this implementation, each kernel contains only half of the operands
 * (x[], d[], m[]) as 15-bit words. Even-numbered kernels contain the
 * low half, odd-numbered kernels contain the high half.
 *
 * Each inner loop execution computes d <- (d + xi*x + f*m) >> 15:
 *
 * -- Loop entry: receives xi, f, input carry, and a delayed upper block
 * -- Loop exit: produces carry and a delayed final block
 *
 * The lowest chunk does not receive a carry or f; it computes f and starts
 * with a carry value of 0.
 * The highest chunk does not receive a delayed upper block; instead, it
 * uses the slot to keep the extra one-bit carry.
 * The value xi is produced by the chunk that contains it.
 */

__kernel __attribute__((reqd_work_group_size(256,1,1))) void
multi_square_2048_15_2(
	__global uint *gx,
	__constant uint *gm,
	uint m0i, int w)
{
	bool low_chunk, high_chunk;
	uint ga, la, lo;
	uint x[69];
	uint d[69];
	uint m[69];
	__local uint cxi[256], cf[256], cr[256], cu[256];

	/*
	 * Read values from global memory. Low half is used by even kernels,
	 * high half by odd kernels.
	 */
	la = get_local_id(0);
	low_chunk = (la & 1) == 0;
	high_chunk = !low_chunk;
	lo = low_chunk ? 0 : 69;
	ga = (get_global_id(0) >> 1) * 138 + lo;
	#pragma unroll 69
	for (int i = 0; i < 69; i ++) {
		x[i] = gx[ga + i];
		m[i] = gm[lo + i];
	}

	m0i &= 0x7FFF;
	for (int k = 0; k < w; k ++) {
		uint dh;
		bool nsub;

		/*
		 * Clear the destination buffer.
		 */
		#pragma unroll 69
		for (int i = 0; i < 69; i ++) {
			d[i] = 0;
		}
		cr[la] = 0;
		cu[la] = 0;
		barrier(CLK_LOCAL_MEM_FENCE);

		dh = 0;
		#pragma unroll 1
		for (int i = 0; i < 139; i ++) {
			uint xi;
			uint f;
			bool do_run;

			/*
			 * Each chunk sends its xi and f values to the
			 * next chunk.
			 */
			xi = cxi[la];
			f = cf[la];
			barrier(CLK_LOCAL_MEM_FENCE);
			if (!high_chunk) {
				cxi[la + 1] = xi;
				cf[la + 1] = f;
			}

			/*
			 * New value xi (for the low chunk) is computed by
			 * the chunk that contains it.
			 */
			if (i < 69) {
				if (low_chunk) {
					cxi[la] = x[i];
				}
			} else if (i < 138) {
				if (high_chunk) {
					cxi[la - 1] = x[i - 69];
				}
			}
			barrier(CLK_LOCAL_MEM_FENCE);
			xi = cxi[la];

			/*
			 * The lowest chunk computes the new multiplier.
			 */
			if (low_chunk) {
				f = mul24(mad24(xi, x[0], d[0]) & 0x7FFF, m0i);
				f &= 0x7FFF;
				cf[la] = f;
			}
			barrier(CLK_LOCAL_MEM_FENCE);
			f = cf[la];

			/*
			 * Only the low chunk runs at iteration 0. Only the
			 * high chunk runs at iteration 138.
			 */
			do_run = !(((i == 0) & high_chunk)
				| ((i == 138) & low_chunk));
			if (do_run) {
				uint r;

				r = cr[la];
				#pragma unroll 35
				for (int j = 0; j < 35; j ++) {
					r = mad24(xi, x[j], d[j]) + r;
					r = mad24(f, m[j], r);
					if (j == 0) {
						if (!low_chunk) {
							cu[la - 1] = r & 0x7FFF;
						}
					} else {
						d[j - 1] = r & 0x7FFF;
					}
					r >>= 15;
				}
				barrier(CLK_LOCAL_MEM_FENCE);
				if (!high_chunk) {
					d[68] = cu[la];
				}
				#pragma unroll 34
				for (int j = 35; j < 69; j ++) {
					r = mad24(xi, x[j], d[j]) + r;
					r = mad24(f, m[j], r);
					d[j - 1] = r & 0x7FFF;
					r >>= 15;
				}
				if (high_chunk) {
					uint z;

					z = dh + r;
					d[68] = z & 0x7FFF;
					dh = z >> 15;
				} else {
					cr[la + 1] = r;
				}
			}
			barrier(CLK_LOCAL_MEM_FENCE);
		}

		/*
		 * Propagate last word for all chunks (except the high one,
		 * which is special).
		 */
		if (!high_chunk) {
			d[68] = cu[la];
		}

		/*
		 * High chunk decides whether there is a subtraction to
		 * perform. The "do it" flag is stored in the cf[] array.
		 */
		if (high_chunk) {
			uint ts;

			ts = dh != 0;
			cf[la] = ts;
			cf[la - 1] = ts;
		}
		cr[la] = 0;
		barrier(CLK_LOCAL_MEM_FENCE);

		/*
		 * Copy and possibly subtract the modulus to ensure that
		 * the value fits. This must be done in several loops
		 * because each chunk needs the carry from the previous
		 * one.
		 */
		nsub = cf[la];
		#pragma unroll 1
		for (int i = 0; i < 2; i ++) {
			if (i == (la & 1)) {
				uint cc;

				cc = cr[la];
				barrier(CLK_LOCAL_MEM_FENCE);
				#pragma unroll 69
				for (int j = 0; j < 69; j ++) {
					uint xj;

					xj = d[j];
					if (nsub) {
						uint yj;

						yj = xj - m[j] - cc;
						cc = (yj >> 15) & 1;
						xj = yj & 0x7FFF;
					}
					x[j] = xj;
				}
				if (!high_chunk) {
					cr[la + 1] = cc;
					barrier(CLK_LOCAL_MEM_FENCE);
				}
			}
		}
	}

	/*
	 * Write back result into global memory.
	 */
	#pragma unroll 69
	for (int i = 0; i < 69; i ++) {
		gx[ga + i] = x[i];
	}
}
